# -*- coding: utf8 -*-
from .Config import Config
from .Downloader import Downloader
from ..utils.utils import *
from ..utils.ColorString import ColorString
from ..core.OptiFine import OptiFine
import os
import uuid
import datetime
import shutil

class Client:

    def __init__(self, config):
        self.config = config
        self.downloader = Downloader(self.config)
    
    def start(self):
        if not self.config.is_client:
            return 

        if self.config.isPure :
            self.downloader.downloadGameJSON()
            self.downloader.downloadClient()
            self.downloader.downloadAssetIndex()
            self.downloader.downloadAssetObjects()
            self.downloader.donwloadLibraries()
            self.writeLauncherProfilesJSON()
        elif self.config.isForge:
            self.config.getForgeInfo()
            if not os.path.exists(self.config.game_version_client_jar_file_path()) or not os.path.exists(self.config.game_version_forge_json_file_path()):
                self.extractForgeClient()
        else:
            ColorString.warn('Not Known Client!!!!')
            return

        user = self.config.username
        resolution = ('960', '540')

        if is_sigint_up():
            return
        
        cmd = self.gameArguments(user,resolution)
        backgroundCmd =  cmd + ' &'
        if platformType() == 'windows':
            backgroundCmd = 'start ' + cmd
            backgroundCmd = cmd
            
        os.chdir(self.config.game_version_client_dir())
        os.system(backgroundCmd)
        print(ColorString.confirm('Start Client Successfully!!!'))
        self.formatOutputClientCmd(backgroundCmd)

    '''如果是调试模式，则输出命令行指令内容'''
    def formatOutputClientCmd(self, cmd):
        if self.config.debug and len(cmd) > 0: 
            for e in cmd.split(' '):
                if 'jar' in e:
                    for c in e.split(self.config.java_class_path_list_separator()):
                        print(c)
                else:
                    print(e)

    '''提取java类相关启动参数'''
    def javaClassPathList(self):

        javaClassPathList = []

        libs = self.config.game_version_json_obj().get('libraries')
        total = len(libs)
        
        index = 0
        for lib in libs: 
            downloads = lib.get('downloads')

            rules = lib.get('rules')
            isContinue = False
            if None != rules:
                for rule in rules:
                    if None != rule:
                        if rule.get('action') == 'disallow':
                            if rule.get('os').get('name') == platformType():
                                isContinue

                        if rule.get('action') == 'allow':
                            allow_os = rule.get('os')
                            if allow_os and allow_os.get('name') != platformType():
                                isContinue = True

            if isContinue:
                continue

            libPath = None
            url = None
            sha1 = None
            nativeKey = 'natives-'+ platformType()
            if 'natives' in lib:
                platform = lib.get('natives').get(platformType())
                if platform == None:
                    continue
                else:
                    libPath = downloads.get('classifiers').get(platform).get('path')
                    url = downloads.get('classifiers').get(platform).get('url')
                    sha1 = downloads.get('classifiers').get(platform).get('sha1')
                    nativeFilePath = os.path.join(self.config.game_version_client_native_library_dir(),os.path.basename(url))
                    if not checkFileExist(nativeFilePath,sha1):
                        continue
                    else:
                        javaClassPathList.append(nativeFilePath)
            else:
                classifiers = downloads.get('classifiers')
                if classifiers and nativeKey in downloads.get('classifiers'):
                    url = downloads.get('classifiers').get(nativeKey).get('url')
                    sha1 = downloads.get('classifiers').get(platform).get('sha1')
                    nativeFilePath = os.path.join(self.config.game_version_client_native_library_dir(),os.path.basename(url))
                    if not checkFileExist(nativeFilePath,sha1):
                        continue
                    else:
                        javaClassPathList.append(nativeFilePath)
                        
                libPath = downloads.get('artifact').get('path')
                url = downloads.get('artifact').get('url')
                sha1 = downloads.get('artifact').get('sha1')
                if platformType() == 'windows' :
                    libPath = libPath.replace('/','\\')
                libFilePath = os.path.join(self.config.game_version_client_library_dir(), libPath)
                if not checkFileExist(libFilePath,sha1):
                    continue            
                else:
                    javaClassPathList.append(libFilePath)

            index = index + 1

        if self.config.isForge:
            forge_class_path = map(lambda lib:  os.path.join(self.config.game_version_client_library_dir(),lib.get('downloads').get('artifact').get('path')) ,self.config.game_version_json_obj().get('libraries')) 
            javaClassPathList.extend(forge_class_path)

        javaClassPathList.append(self.config.game_version_client_jar_file_path())

        return javaClassPathList


    def gameArguments(self, user, resolution):

        argPattern = r'\$\{(.*)\}'

        mainCls =  self.config.game_version_json_obj().get('mainClass')
        classPathList = self.javaClassPathList()

        if self.config.isPure and self.config.optifine:
            optifine_config = OptiFine.json_configuration(self.config)
            if optifine_config != None:
                mainCls = optifine_config.get('mainClass')
                optifine_jar_paths = OptiFine.library_optifine_jar_paths(self.config)
                classPathList = optifine_jar_paths + classPathList

        sep = self.config.java_class_path_list_separator()
        classPath = sep.join(classPathList)
        (res_width, res_height) = resolution

        configuration = {
            # for game args
            "auth_player_name" : user,
            "version_name" : self.forgeGame().get('id') if self.config.isForge else self.config.game_version_json_obj().get('id'),
            "game_directory" : self.config.game_version_client_dir(),
            "assets_root" : self.config.game_version_client_assets_dir(),
            "assets_index_name" : self.config.game_version_json_obj().get('assets'),
            "auth_uuid" : ''.join(str(uuid.uuid1()).split('-')),
            "auth_access_token" : ''.join(str(uuid.uuid1()).split('-')),
            "user_type" : "mojang",
            "version_type" : "release",
            "resolution_width":res_width if res_width else "",
            "resolution_height":res_height if res_height else "",
            # for jvm args
            "natives_directory": self.config.game_version_client_native_library_dir(),
            "launcher_name": "OrzMC",
            "launcher_version": '1.0',
            "classpath": classPath,

        }

        java_path = os.popen('which java' if platformType() != 'windows' else 'where java').read().strip()
        if ' ' in java_path:
            java_path = '\"' + java_path + '\"'

        if len(java_path) <= 0:
            ColorString.error('You should installl JDK to run Minecraft')
            exit(-1)

        java_path = 'java'

        arguments = [java_path]

        mem_min = self.config.mem_min
        mem_max = self.config.mem_max

        jvm_opts = [
            '-Xms%s' % mem_min,
            '-Xmx%s' % mem_max,
            '-Djava.net.preferIPv4Stack=true'
        ]
        arguments.extend(jvm_opts)

        game_arguments = self.config.game_version_json_obj().get('arguments')
        jvmArgs = game_arguments.get('jvm') if game_arguments != None else None

        if jvmArgs != None and self.config.isForge:
            jvmArgs.append('-Dfml.ignoreInvalidMinecraftCertificates=true')
            jvmArgs.append('-Dfml.ignorePatchDiscrepancies=true')

        if jvmArgs != None:
            for arg in jvmArgs:
                if isinstance(arg, str):
                    value_placeholder = re.search(argPattern,arg)
                    if value_placeholder:
                        argValue = configuration.get(value_placeholder.group(1))
                        argStr = matchAndReplace(argPattern,argValue,arg)
                        arguments.append(argStr)
                    else:
                        arguments.append(arg)

                elif isinstance(arg, dict):
                    isValid = False
                    rules = arg.get('rules')
                    for rule in rules:
                        if rule.get('os').get('name') == platformType() and rule.get('action') == 'allow':
                            isValid = True
                            break
                    if isValid:
                        value = arg.get('value')
                        if type(value) == list:
                            for i in range(0, len(value)):
                                s = value[i].split('=')
                                if len(s) == 2:
                                    arg = s[1]
                                    if ' ' in arg:
                                        s[1] = '\"' + arg + '\"'
                                        value[i] = '='.join(s)

                            arguments.extend(value)
                        else:
                            arguments.append(value)


        arguments.append(mainCls)

        gameArgs = game_arguments.get('game') if game_arguments != None else None

        if gameArgs != None:
            for arg in gameArgs:
                if isinstance(arg, str):
                    value_placeholder = re.search(argPattern,arg)
                    if value_placeholder:
                        argValue = configuration.get(value_placeholder.group(1))
                        argStr = matchAndReplace(argPattern,argValue,arg)
                        arguments.append(argStr)
                    else:
                        arguments.append(arg)
                elif isinstance(arg, dict):
                    for rule in arg.get('rules'):
                        if rule.get('action') == 'allow':
                            if rule.get('features').get('is_demo_user'):
                                if user ==  None:
                                    arguments.append(arg.get('value'))
                            if rule.get('features').get('has_custom_resolution'):
                                value = arg.get('value')
                                for arg in value:
                                    if isinstance(arg, str):
                                        value_placeholder = re.search(argPattern,arg)
                                        if value_placeholder:
                                            argValue = configuration.get(value_placeholder.group(1))
                                            argStr = matchAndReplace(argPattern,argValue,arg)
                                            arguments.append(argStr)
                                        else:
                                            arguments.append(arg)

        if self.config.isForge:
            arguments.extend(self.config.game_version_json_obj().get('arguments').get('game'))
        
        if self.config.isPure and self.config.optifine:
            optifine_config = OptiFine.json_configuration(self.config)
            if optifine_config != None:
                arguments.extend(optifine_config.get('arguments').get('game'))
        arguments = ' '.join(arguments)
        return arguments

    '''构建Forge客户端'''
    def extractForgeClient(self):
        
        self.download(
            self.config.forgeInfo.forge_installer_url, 
            self.config.game_version_client_dir(), 
            prefix_desc='forge installer jar file'
        )

        installerJarFilePath = os.path.basename(self.config.forgeInfo.forge_installer_url)
        extractForgeClientCmd = 'java -jar ' + installerJarFilePath

        print(ColorString.warn('Start install the forge client jar file ...'))
        os.chdir(self.config.game_version_client_dir())
        os.system(extractForgeClientCmd)
        print(ColorString.confirm('Completed! And the forge client file generated!'))

        forge_client_versions_dir = os.path.join(self.config.game_version_client_dir(), 'versions')
        pure_client_jar = os.path.join(forge_client_versions_dir, self.config.version, self.config.version + '.jar')
        dst_client_jar = os.path.join(self.config.game_version_client_dir(), os.path.basename(pure_client_jar))
        if os.path.exists(pure_client_jar) and not os.path.exists(dst_client_jar):
            shutil.move(pure_client_jar, self.config.game_version_client_dir())

        forge_version_full_name = os.path.basename(self.config.game_version_forge_json_file_path())
        (forge_version_name, _) = os.path.splitext(forge_version_full_name)
        forge_json_file = os.path.join(forge_client_versions_dir, forge_version_name, forge_version_full_name)
        print(forge_json_file)
        dst_forge_json_file = os.path.join(self.config.game_version_client_dir(), os.path.basename(forge_json_file))
        if os.path.exists(forge_json_file) and not os.path.exists(dst_forge_json_file):
            shutil.move(forge_json_file, self.config.game_version_client_dir())

        shutil.rmtree(forge_client_versions_dir)

    def writeLauncherProfilesJSON(self):
        launcher_profiles_json_file_path = self.config.game_version_launcher_profiles_json_path()
        if not os.path.exists(launcher_profiles_json_file_path):
            time = datetime.datetime.now().isoformat()
            version = self.config.version
            content = {
                'profiles': { 
                    version: {
                        'created': time,
                        'type': 'pure',
                        'name': version,
                        'lastVersionId': version,
                        'lastUsed': time
                    }
                },
                'selectedProfile': version,
                'authenticationDatabase': {}
            }

            launcher_profiles_json_file_content = json.dumps(content)

            with open(launcher_profiles_json_file_path,'w',encoding='utf-8') as f:
                f.write(launcher_profiles_json_file_content)


# -*- coding: utf8 -*-

import hashlib
import os
import sys
import platform
import zipfile
import re
import json

def matchAndReplace(pattern, repl, string):

    if platformType() == 'windows':
        repl = repl.replace('\\', '@')

    ret = re.sub(pattern, repl, string)

    if platformType() == 'windows':
        ret = ret.replace('@', '\\')
        
    return ret

def checkFileExist(filePath, hash):
    if not os.path.exists(filePath):
        return False
    ret = (computeHash(filePath) == hash)   
    if not ret:
        # print('sha1 check failed: ' + filePath)
        pass
    return ret

def computeHash(filePath):
    if not os.path.exists(filePath):
        return None
    with open(filePath, 'rb') as f:
        computeHash = hashlib.sha1(f.read()).hexdigest()
        return computeHash

def writeContentToFile(content, filePath):
        if content != None:
            with open(filePath,'w',encoding='utf-8') as f:
                f.write(content)

def makedirs(path):
    if not os.path.exists(path):
        os.makedirs(path)


def platformType():
    '''OS type'''
    system = {
        'Linux':'linux',
        'Darwin':'osx',
        'Windows': 'windows'
    }
    return system[platform.system()]

def hint(msg):
    return input(msg)

def zip(srcPaths, dstPath):
    if len(srcPaths) <= 0:
        return
    try:
        with zipfile.ZipFile(dstPath, 'w', zipfile.ZIP_DEFLATED, allowZip64=True) as myzip:
            for srcPath in srcPaths:
                if os.path.exists(srcPath):
                    if os.path.isdir(srcPath):
                        for dirpath, _, filenames in os.walk(srcPath):
                            dir_path = srcPath.replace(os.path.basename(srcPath), '')
                            fpath = dirpath.replace(dir_path, '')
                            fpath = fpath and fpath + os.sep or ''
                            for filename in filenames:
                                source = os.path.join(dirpath,filename)
                                destination = fpath + filename
                                myzip.write(source, destination)
                    else:
                        myzip.write(srcPath)
    except Exception as e:
        print(e)


def loadJSON(filePath):
    with open(filePath) as json_data:
        jsonObj = json.load(json_data)
        return jsonObj

from .CleanUp import CleanUp
def is_sigint_up():
    return CleanUp.is_sigint_up

def convertOggToMap3(source_file_path, target_file_path):
    cmd = 'ffmpeg -i ' + source_file_path + ' ' + target_file_path + ' -y'
    os.system(cmd)  


# -*- coding: utf8 -*-
from ..core.Oracle import Oracle
from ..utils.ColorString import ColorString
from ..utils.utils import *
from .Config import Config
import argparse
import re
import os

def install_jdk():
    Oracle.install_jdk()

def uninstall_jdk():
    Oracle.execute_uninstall_jdk()

def rsync_server_core_data():
    '''服务器迁移时用来迁移服务器核心数据'''
    #命令行参数解析 argparse 使用文档：https://docs.python.org/3/library/argparse.html
    parser = argparse.ArgumentParser(description='use rsync command to sync the minecraft server core data to other location or host')

    # Tool
    parser.add_argument('-s', metavar='source', dest='source', help='Specified the source file or dir to sync')
    parser.add_argument('-d', metavar='destination', dest='destination', help='Specified the destination dir to sync')
    parser.add_argument('-y', '--yes', default=False, action='store_true', help='ask yes when require user select')
    args = parser.parse_args()
    
    source = args.source
    destination = args.destination

    ftp_server_base_dir_name = os.path.basename(Config.game_ftp_server_base_dir())
    server_core_data_dir_name = os.path.basename(Config.game_ftp_server_core_data_backup_dir())

    server_core_data_dir_path = os.path.join(os.path.expanduser('~'),"%s/%s" % (ftp_server_base_dir_name, server_core_data_dir_name))
    if not source and os.path.exists(server_core_data_dir_path):
        source = server_core_data_dir_path
        if os.path.isdir(server_core_data_dir_path):
            source += '/*' 

    def check_args(source, destination):
        if not destination or not source:
            print(ColorString.warn('You should provide both source and destination argument for this command, destination can be a (local dir/file) remote host (example: ubuntu@mc.jokerhub.cn)'))
            exit(-1)

    def execute_sync(source, destination, test = True):
        check_args(source, destination)

        pattern = re.compile(r'\w+@\w+')
        dest = destination.strip()
        source = source.strip()
        match = re.match(pattern, dest)
        if match:
            ftp_server_base_dir_name = os.path.basename(Config.game_ftp_server_base_dir())
            
            sync_file_dir_name = os.path.basename(source)
            if not os.path.exists(source):
                segments = list(os.path.split(source))[0:-1]
                sync_file_dir_name = os.path.basename(os.path.join(*segments))
            dest += ':~/%s/%s' % (ftp_server_base_dir_name,sync_file_dir_name)

        rsync_cmd = 'rsync -zarv %s %s ' % (source, dest)
        rsync_cmd += "--exclude 'plugins/dynmap/*'"
        if test:
            rsync_cmd += ' -n'

        os.system(rsync_cmd)

        if test:
            print('\ncommand: %s' % ColorString.confirm(rsync_cmd))
            print(ColorString.hint("Run in Fake Mode!"))

    check_args(source = source, destination = destination)
    execute_sync(source = source, destination = destination, test = True)

    confirm = ['Y','y','Yes','yes']
    cancel = ['N','n','No','no']
    while True:
        a = hint(ColorString.confirm('\nAre you confirm want to execute this operation? [%s] ' % ('/'.join(confirm) + '|' + '/'.join(cancel))))
        if a in confirm:
            execute_sync(source=source, destination = destination, test = False)
            break
        elif a in cancel:
            break
        else:
            print(ColorString.warn('Your input is invalid, Please try again!'))




# -*- coding: utf8 -*-

import argparse
from .Config import Config
from .Game import Game

# argparse 使用文档：https://docs.python.org/3/library/argparse.html
def parse_args():
    '''命令行参数解析'''
    parser = argparse.ArgumentParser(description='A command line tool for start Minecraft client or deploy minecraft server')

    # Tool
    parser.add_argument('-v', metavar='version', dest='version', help='Specified the Minecraft clinet version number to start')

    # Minecraft
    parser.add_argument('-u', metavar='username', default=Config.GAME_DEFAULT_USERNAME, dest='username', help='pick an username(default: %(default)s) for player when start the client')
    server_type_choices = [Config.GAME_TYPE_PURE, Config.GAME_TYPE_FORGE, Config.GAME_TYPE_SPIGOT, Config.GAME_TYPE_PAPER]
    client_type_choices = [Config.GAME_TYPE_PURE, Config.GAME_TYPE_FORGE]
    type_choices = list(set(server_type_choices + client_type_choices))
    sep = '/'
    type_help_info = 'Specified the type of game: "%s" for server, "%s" for client' % (sep.join(server_type_choices), sep.join(client_type_choices))
    parser.add_argument('-t', default=Config.GAME_TYPE_PURE, choices=type_choices, metavar='type', dest='type', help=type_help_info)
    
    # JVM
    parser.add_argument('-m', default='512M', metavar='minmem', dest='minmem', help='Specified the JVM initial memory allocation(default: %(default)s)')
    parser.add_argument('-x', default='2G', metavar='maxmem', dest='maxmem', help='Specified the JVM max memory allocation(default:%(default)s)')

    # Flags
    parser.add_argument('-s','--server', default=False, action='store_true', help='deploy minecraft server, if there is no this flag, this command line tool start minecraft as client')
    parser.add_argument('-f','--force_upgrade_world', default=False, action='store_true', help='when deploy spigot/paper server, the option can upgrade your map from old game version')
    parser.add_argument('-b','--backup_world', default=False, action='store_true', help='backup your minecraft world into ~/minecraft_world_backup directory as zip file!')
    parser.add_argument('-o','--optifine', default=False, action='store_true', help='if you have installed optifine for client, you can add this option to launch client with optifine be activated')
    parser.add_argument('-D','--debug', default=False, action='store_true', help='Run Command in Debug Mode')
    parser.add_argument('-V','--verbose', default=False, action='store_true', help='Output some debug info for bugfix')

    # server options
    parser.add_argument('--api', default = 'v2', metavar = 'api', dest='api', help = 'select paper api version(v1/v2) to download server jar file, default is: v2')
    parser.add_argument('--force_download', default=False, action = 'store_true', help= 'force download server jar file ignore existed!')
    parser.add_argument('-l', '--symlink', default=False, action='store_true',help='create symlink files for current version server core file, and make server version upgrade easy.')

    # extract bgm music
    parser.add_argument('-e', '--extract_music', default=False, action='store_true', help='extract specific version client music')

    args = parser.parse_args()

    return args


def start():
    '''启动游戏'''
    # 控制台收集的参数传入Config对象进行初始化
    args = parse_args()
    config = Config(args)
    Game(config).start()



# -*- coding: utf8 -*-

from os.path import isfile
from .Config import Config
from .Downloader import Downloader
from ..utils.ColorString import ColorString
from ..utils.CleanUp import CleanUp
from ..utils.utils import *

from ..core.Spigot import Spigot

import os
import io
import time
import shutil 

class Server:
    def __init__(self, config):
        self.config = config
        self.downloader = Downloader(self.config)
        
    '''start minecraft server'''
    def start(self):
        
        if self.config.is_client:
            return

        if self.config.backup:
            self.backupWorld()
            return
        
        if self.config.isPure:
            self.downloader.downloadGameJSON()
            self.downloader.downloadServer()
            (serverJARFilePath, _, _) = self.serverJARFilePath()
            jarFilePath = serverJARFilePath
        elif self.config.isSpigot:
            if self.config.force_download or not os.path.exists(self.config.game_version_server_jar_file_path()):
                self.buildSpigotServer()
            jarFilePath = self.config.game_version_server_jar_file_path()
        elif self.config.isPaper:
            if self.config.force_download or not os.path.exists(self.config.game_version_server_jar_file_path()):
                self.downloader.downloadPaperServerJarFile()
            jarFilePath = self.config.game_version_server_jar_file_path()
        elif self.config.isForge:
            self.config.getForgeInfo()
            if self.config.force_download or not os.path.exists(self.config.game_version_server_jar_file_path()):
                self.buildForgeServer()
            jarFilePath = self.config.game_version_server_jar_file_path()
            if not os.path.exists(jarFilePath):
                jarFilePath = jarFilePath.replace(self.config.forgeInfo.fullVersion, self.config.forgeInfo.fullVersion + '-universal')
        else:
            print(ColorString.warn('Your choosed server is not exist!!!\nCurrently, there are three type server: pure/spigot/forge'))
            return

        jvm_opts = ' '.join([
            '-server',
            '-Xms' + self.config.mem_min,
            '-Xmx' + self.config.mem_max
        ])
        jarArgs = ['--forceUpgrade', 'nogui'] if (self.config.isSpigot or self.config.isPaper) and self.config.force_upgrade else ['nogui']
        self.startServer(self.startCommand(jvm_opts= jvm_opts, serverJARFilePath = jarFilePath, jarArgs = jarArgs))
        print(ColorString.confirm('Start Server Successfully!!!'))
    
    '''服务端运行需要的JAR文件所在路径'''
    def serverJARFilePath(self):
            server = self.config.game_version_json_obj().get('downloads').get('server')
            serverUrl = server.get('url')
            sha1 = server.get('sha1')
            serverJARFilePath = self.config.game_version_server_jar_file_path()
            return (serverJARFilePath, serverUrl, sha1)
    
    '''Download Server Jar File'''
    def downloadServer(self):    
        (serverJARFilePath, serverUrl, sha1) = self.serverJARFilePath()
        if not checkFileExist(serverJARFilePath, sha1):
            self.downloader.download(
                serverUrl,
                self.config.game_version_server_dir(), 
                self.config.game_version_server_jar_filename(), 
                prefix_desc='server jar file'
            )
        else:
            print("server jar file existed!")

    '''construct server start command'''
    def startCommand(self, jvm_opts = '', serverJARFilePath = '', jarArgs = ['nogui']):
        argList = [
            'java',
            jvm_opts,
            '-jar',
            serverJARFilePath
        ]
        argList.extend(jarArgs)
        cmd = ' '.join(argList)
        return cmd

    def checkEULA(self):
        return os.path.exists(self.config.game_version_server_eula_file_path())

    '''启动minecraft服务器'''
    def startServer(self, cmd):

        os.chdir(self.config.game_version_server_dir())

        # 如果没有eula.txt文件，则启动服务器生成
        if not self.checkEULA():
            if self.config.isForge:
                self.generateForgeServerEULA()
            else:
                os.system(cmd)

        # 同意eula
        with io.open(self.config.game_version_server_eula_file_path(), 'r', encoding = 'utf-8') as f:
            eula = f.read()
            checkEULA = eula.replace('false', 'true')
        with io.open(self.config.game_version_server_eula_file_path(), 'w', encoding = 'utf-8') as f:
            f.write(checkEULA)
        
        # 启动服务器
        
        if self.config.debug:
            print(cmd)

        os.system(cmd)

        # 设置服务器属性为离线模式
        with io.open(self.config.game_version_server_properties_file_path(), 'r', encoding = 'utf-8') as f:
            properties = f.read()
            if 'online-mode=false' in properties:
                offline_properties = None
            else:
                offline_properties = properties.replace('online-mode=true', 'online-mode=false')

        if offline_properties != None:
            with io.open(self.config.game_version_server_properties_file_path(), 'w', encoding = 'utf-8') as f:
                f.write(offline_properties)
                print(ColorString.confirm('Setting the server to offline mode, next launch this setting take effect!!!'))

        # 为服务器核心数据文件创建符号链接
        self.symlink_server_core_files_if_need()


    # 构建SpigotServer
    def buildSpigotServer(self):
        '''构建SpigotServer'''

        version = self.config.version
        spigot = Spigot(version)
        self.download(
            spigot.build_tool_jar, 
            self.config.game_version_server_build_dir(), 
            prefix_desc='spigot build tool jar file'
        )
        buildToolJarName = os.path.basename(spigot.build_tool_jar)
        buildToolJarPath = os.path.join(self.config.game_version_server_build_dir(), buildToolJarName)
        versionCmd = ' --rev ' + version if len(version) > 0 else 'lastest'
        linuxCmd = 'git config --global --unset core.autocrlf; java -jar ' + buildToolJarPath + versionCmd
        macCmd = 'export MAVEN_OPTS="-Xmx2G"; java -Xmx2G -jar ' + buildToolJarPath + versionCmd
        windowCmd = 'java -jar ' + buildToolJarPath + versionCmd
        cmd = macCmd if platformType() == 'osx' else linuxCmd if platformType() == 'linux' else windowCmd
        print(ColorString.warn('Start build the server jar file with build tool...'))
        os.chdir(self.config.game_version_server_build_dir())
        os.system(cmd)
        print(ColorString.confirm('Completed! And the spigot built server file generated!'))
        shutil.move(self.config.game_version_server_jar_file_path(isInBuildDir=True), self.config.game_version_server_jar_file_path())
        os.chdir(self.config.game_version_server_dir())
        shutil.rmtree(self.config.game_version_server_build_dir())
    
    # 构建Forge服务器
    def buildForgeServer(self):
        '''构建Forge服务器'''

        self.download(
            self.config.forgeInfo.forge_installer_url, 
            self.config.game_version_server_dir(), 
            prefix_desc='forge installer jar file'
        )

        installerJarFilePath = os.path.basename(self.config.forgeInfo.forge_installer_url)
        installServerCmd = 'java -jar ' + installerJarFilePath + ' --installServer'
        print(ColorString.warn('Start install the forge server jar file ...'))
        os.chdir(self.config.game_version_server_dir())
        os.system(installServerCmd)
        print(ColorString.confirm('Completed! And the forge server file generated!'))


    def generateForgeServerEULA(self):
        if self.config.isForge and not self.checkEULA(): 
            pure_server_jar = os.path.join(self.config.game_version_server_dir(), '.'.join(['minecraft_server',self.config.version,'jar']))
            boot_cmd = 'java -jar ' + pure_server_jar
            os.system(boot_cmd)

    def backupWorld(self):
        world_paths = self.config.game_version_server_world_dirs()
        if world_paths:            
            backup_path = self.config.game_version_server_world_backup_dir()
            now = time.localtime()
            fileName = '_'.join([time.strftime('%Y-%m-%dT%H:%M:%S', now), self.config.game_type, self.config.version]) + '.zip'
            world_backup_file = os.path.join(backup_path, fileName)

            def backupWorld_cleanUp():
                if os.path.isfile(world_backup_file) and os.path.exists(world_backup_file):
                    os.remove(world_backup_file)
                    print(ColorString.warn("Removed Invalid World Backup File: %s" % world_backup_file))

            print(ColorString.hint('Start Executing ZIP ...'))
            CleanUp.registerTask('backupWorld_cleanUp', backupWorld_cleanUp)
            zip(world_paths, world_backup_file)
            CleanUp.cancelTask('backupWorld_cleanUp')
            file_size = os.path.getsize(world_backup_file) / 1024.0 / 1024.0 / 1024.0
            print(ColorString.confirm("Completed! backuped world file(%.2fG): %s!!!" % (file_size, world_backup_file)))
        else: 
            print(ColorString.error('There is no world directory!!!'))

    def symlink_server_core_files_if_need(self):
        '''为服务器核心文件创建符号链接到共享目录，方便版本升级'''
        if self.config.symlink:
            # 软链接需要涉及的文件和目录
            destinations = [
                self.config.game_version_server_eula_file_path(),
                self.config.game_version_server_properties_file_path(),
                self.config.game_version_server_icon_file_path(),
                self.config.game_version_server_plugin_dir(),
            ]
            world_paths = self.config.game_version_server_world_dirs()
            if world_paths and len(world_paths) > 0:
                destinations.extend(self.config.game_version_server_world_dirs())

            # 遍历创建软链接
            for destination in destinations:
                # 软链接需要链接的源文件或源目录
                source = os.path.join(self.config.game_version_server_symlink_source_dir(),os.path.basename(destination))
                # 如果软链接生成的目录下有同名的文件和目录，则取个备份用的名称
                destination_backup = destination+'.before_symlink'

                if not os.path.exists(source):
                    if os.path.exists(destination):
                        shutil.move(destination, source)
                    else:
                        # 之前有备份过的同名文件和目录，则恢复同名的文件和目录
                        if os.path.exists(destination_backup):
                            shutil.move(destination_backup, destination)
                        continue

                # 如果已经创建了软链接，则跳过不再创建
                if os.path.islink(destination):
                    target = os.readlink(destination)
                    if os.path.normcase(target) == os.path.normcase(source):
                        continue
                
                # 如果软链接生成的目录下有同名文件或目录，则先进行备份
                if os.path.exists(destination):
                    shutil.move(destination, destination_backup)
                
                # 生成软链接
                os.symlink(source, destination, target_is_directory=os.path.isdir(source))
                print("%s -> %s" % (source, destination))


# -*- coding: utf8 -*-
from ..utils.utils import *
from ..utils.ColorString import ColorString
from ..core.PaperAPI import PaperAPI
from ..core.Mojang import Mojang
from ..core.Spigot import Spigot
from ..core.Forge import Forge
from ..core.OptiFine import OptiFine

from tqdm import tqdm
import requests
import shutil

class Downloader:

    def __init__(self, config):
        self.config = config

    '''Download Game Json Configure File'''
    def downloadGameJSON(self):

        if is_sigint_up():
            return

        version_json_path = self.config.game_version_json_file_path()
        (url, hash) = Mojang.get_release_game_json(self.config.version)
        if not checkFileExist(version_json_path, hash):
            print("Download Game Json Configure File!")
            jsonStr = requests.get(url).text
            writeContentToFile(jsonStr, version_json_path)
        else: 
            print("Game Json Configure File have been downloaded!")

    def downloadClient(self):
        if is_sigint_up():
            return  

        '''Download Client Jar File'''
        client = self.config.game_version_json_obj().get('downloads').get('client')
        clientUrl = client.get('url')
        sha1 = client.get('sha1')
        clientJARFilePath = self.config.game_version_client_jar_file_path()
        if not checkFileExist(clientJARFilePath, sha1):
            self.download(
                clientUrl,
                os.path.dirname(clientJARFilePath), 
                self.config.game_version_client_jar_filename(), 
                prefix_desc='client jar file'
            )
        else:
            print("client jar file existed!")

    '''Download Game Asset Index JSON file'''
    def downloadAssetIndex(self):

        if is_sigint_up():
            return

        assetIndex = self.config.game_version_json_obj().get('assetIndex')
        index_json_url = assetIndex.get('url')
        index_json_sha1 = assetIndex.get('sha1')
        index_json_path= os.path.join(self.config.game_version_client_assets_indexs_dir(), os.path.basename(index_json_url))
        if not checkFileExist(index_json_path,index_json_sha1):
            print("Download assetIndex JSON File")
            index_json_str = requests.get(index_json_url).text
            with open(index_json_path,'w',encoding='utf-8') as f:
                f.write(index_json_str)
        else:
            print("assetIndex JSON File have been downloaded")
    
    '''Download Game Asset Objects'''
    def downloadAssetObjects(self):

        objects = self.config.game_version_json_assets_obj().get('objects')
        total = len(objects)

        index = 0
        for (name,object) in objects.items():

            if is_sigint_up():
                return

            index = index + 1
            hash = object.get('hash')
            url = Mojang.assets_objects_url(hash)
            object_dir = self.config.game_version_client_assets_objects_dir(hash)
            object_filePath = os.path.join(object_dir,os.path.basename(url))
            if not checkFileExist(object_filePath, hash):
                prefix_desc = 'assets objects %d/%d(%.2f%%)' % (index, total, 100.0 * index / total)
                self.download(url,object_dir, prefix_desc=prefix_desc)
            
            # 收集ogg音乐文件
            if self.config.is_extract_music:
                path, type = os.path.splitext(name)
                filename = path.replace(os.sep, '_') + '.mp3'
                if type == '.ogg':
                    target_file_path = os.path.join(self.config.game_version_client_mp3_dir(),filename)
                    convertOggToMap3(object_filePath, target_file_path)

        if self.config.is_extract_music:
            music_dir = os.path.dirname(self.config.game_version_client_mp3_dir())
            print(ColorString.confirm('music has been extracted to dir: %s' % music_dir))
            exit(0)

    ''' download libraries'''
    def donwloadLibraries(self):

        libs = self.config.game_version_json_obj().get('libraries')
        total = len(libs)

        index = 0
        for lib in libs: 

            if is_sigint_up():
                return

            index = index + 1
            prefix_desc = 'libraries %d/%d' % (index, total)
            downloads = lib.get('downloads')
            rules = lib.get('rules')

            isContinue = False
            if None != rules:
                for rule in rules:
                    if None != rule:
                        if rule.get('action') == 'disallow':
                            if rule.get('os').get('name') == platformType():
                                isContinue = True

                        if rule.get('action') == 'allow':
                            allow_os = rule.get('os')
                            if allow_os and allow_os.get('name') != platformType():
                                isContinue = True

            if isContinue: 
                continue

            libPath = None
            url = None
            nativeKey = 'natives-'+ platformType()
            if 'natives' in lib:
                platform = lib.get('natives').get(platformType())
                if platform == None:
                    continue
                else:
                    libPath = downloads.get('classifiers').get(platform).get('path')
                    url = downloads.get('classifiers').get(platform).get('url')
                    sha1 = downloads.get('classifiers').get(platform).get('sha1')
                    nativeFilePath = os.path.join(self.config.game_version_client_native_library_dir(),os.path.basename(url))
                    if not checkFileExist(nativeFilePath,sha1):
                        self.download(url,self.config.game_version_client_native_library_dir(), prefix_desc=prefix_desc)
                    
            else:
                classifiers = downloads.get('classifiers')
                if classifiers and nativeKey in downloads.get('classifiers'):
                    url = downloads.get('classifiers').get(nativeKey).get('url')
                    sha1 = downloads.get('classifiers').get(platform).get('sha1')
                    nativeFilePath = os.path.join(self.config.game_version_client_native_library_dir(),os.path.basename(url))
                    if not checkFileExist(nativeFilePath,sha1):
                        self.download(url,self.config.game_version_client_native_library_dir(), prefix_desc=prefix_desc)
                
                libPath = downloads.get('artifact').get('path')
                url = downloads.get('artifact').get('url')
                sha1 = downloads.get('artifact').get('sha1')
                fileDir = self.config.game_version_client_library_dir(libPath)
                filePath=os.path.join(fileDir,os.path.basename(url))
                if not checkFileExist(filePath,sha1):
                    self.download(url,fileDir, prefix_desc=prefix_desc)

    '''下载Paper服务端JAR文件'''
    def downloadPaperServerJarFile(self):
        url = None
        if self.config.api == 'v1':
            url = PaperAPI.downloadURLV1(
                project_name = 'paper',
                project_version = self.config.version,
                build_id = 'latest'
            )
        elif self.config.api == 'v2':
            url = PaperAPI.downloadURLV2(
                version = self.config.version
            )

        if url and len(url) > 0:
            self.download(
                url, 
                self.config.game_version_server_dir(), 
                self.config.game_version_server_jar_filename(), 
                prefix_desc='paper server file'
            )


    '''通用下载方法'''
    def download(self, url, dir, name = None, prefix_desc = None):

        if is_sigint_up():
            return
        if name == None:
            filename = os.path.basename(url)
        else:
            filename = name
        
        try:
            # 临时文件中转目录
            target_file = os.path.join(dir,filename)
            download_temp_file = os.path.join(self.config.game_download_temp_dir(), filename)
            
            total_size = int(requests.head(url).headers["Content-Length"])
            kb_chunk_size = 1024 # 单位: 1K
            mb_chunk_size = 1024 * 1024 # 单位: 1M
            
            # 大于10M的通过流式下载
            if total_size > 10 * mb_chunk_size:
                mb_size = int(total_size / mb_chunk_size + 0.5)
                res = requests.get(url, stream = True)
                # 先下载到临时目录
                with open(download_temp_file,'wb') as f:
                    desc = ((prefix_desc + ' - ') if len(prefix_desc) > 0 else  '') + 'downloading: %s(%sMB)' % (filename, mb_size)
                    for chunk in tqdm(iterable=res.iter_content(mb_chunk_size),total=mb_size,unit='MB',desc=desc):
                        f.write(chunk)

                # 下载并写入临时目录后，移动到目标位置, 如果目标位置已存在文件，先删除
                if os.path.exists(target_file):
                    os.remove(target_file)
                # 移动临时文件到目标位置
                shutil.move(download_temp_file, target_file)
            else:
                # 小于10M的直接下载到内存，然后转存
                with open(target_file, 'wb') as f:
                    f.write(requests.get(url).content)
                kb_size = int(total_size / kb_chunk_size + 0.5)
                desc = (prefix_desc + ' - ' if len(prefix_desc) > 0 else '') + ('%s(%sKB)' % (os.path.basename(url), kb_size))
                print(desc)
        except Exception as e:
            # 如果下载失败, 则提示
            print(e)
            print(ColorString.error('download failed: %s' % url))
            exit(-1)




# -*- coding: utf-8 -*-
import setuptools
import sys

setuptools.setup(
    name="OrzMC",
    version="1.0.46",
    description="A project for deployment of minecraft game.",
    long_description=open('README.md', encoding = 'UTF-8').read(),
    long_description_content_type="text/markdown",
    author="王志舟<wangzhizhou>",
    author_email="824219521@qq.com",
    maintainer='王志舟',
    maintainer_email='824219521@qq.com',
    url="https://github.com/OrzGeeker/OrzMC.git",
    packages=setuptools.find_packages(),
    platforms = ['all'],
    keywords='minecraft python',
    classifiers=(
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux"
    ),
    python_requires='>=3, <4',
    install_requires=[
        'pylint',
        'requests',
        'twine',
        'beautifulsoup4'
    ],
    entry_points = {
        'console_scripts': [
            'orzmc = OrzMC:start',
            'orzmc-rsync = OrzMC:rsync_server_core_data',
            'orzmc-jdk = OrzMC:install_jdk',
            'orzmc-jdk-remove = OrzMC:uninstall_jdk'
        ]
    }
)



#!/usr/bin/env python3
#-*- coding: utf-8 -*-

import os
# os.environ['KIVY_TEXT'] = 'pil'
from kivy.app import App
from kivy.core.window import Window
Window.size = (480, 270)

from kivy.uix.dropdown import DropDown
from kivy.uix.button import Button
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout

from kivy.graphics import Color, Rectangle

from OrzMC.Mojang import Mojang
from OrzMC.Config import Config
from OrzMC.Game import Game


class MCInfoWidget(BoxLayout):
    def __init__(self, **kwargs):
        super(MCInfoWidget,self).__init__(**kwargs)

        versionsDropDownList = DropDown()
        release_versions = Mojang.get_release_version_id_list(update = True)
        for version in release_versions:
            item = Button(
                text = version,
                size_hint_y = None, 
                height = 50
            )
            item.bind(on_release = lambda item: versionsDropDownList.select(item.text))
            versionsDropDownList.add_widget(item)

        self.versionsButton = Button(
            text = release_versions[0] if len(release_versions) > 0 else 'Get Versions Failed!',
            size_hint = (0.3, 1)
        )
        self.versionsButton.bind(on_release = versionsDropDownList.open)
        versionsDropDownList.bind(on_select = lambda item, text: setattr(self.versionsButton, 'text', text))
        self.add_widget(self.versionsButton)

        self.usernameTextInput = TextInput(
            multiline = False,
            hint_text = 'input a username',
            size_hint = (0.7, 1)
        )
        self.add_widget(self.usernameTextInput)

class Content(FloatLayout):
    def __init__(self, **kwargs):
        super(Content, self).__init__(**kwargs)
        
        self.mcInfoWidget = MCInfoWidget(
            pos_hint = {'x': 0, 'y': .8 },
            size_hint_y = None,
            height = 90,
            spacing = 20,
            padding = 20   
        )
        self.add_widget(self.mcInfoWidget)
        
        startButton = Button(
            text = 'Start',
            size = (140, 60),
            size_hint = (None, None),
            pos_hint = {'center_x': .5, 'center_y': 0.1}
        )
        startButton.bind(on_release = self.startGame)
        self.add_widget(startButton)

        self.infoLabel = Label(
            pos_hint = {'center_x': .5, 'center_y': .5},
            size_hint_y = None
        )
        self.add_widget(self.infoLabel)

    def startGame(self, instance):
        version = self.mcInfoWidget.versionsButton.text
        username = self.mcInfoWidget.usernameTextInput.text

        if len(version) > 0 and len(username) > 0 :
            config = Config(
                is_client = True,
                version = version,
                username = username,
                game_type = 'pure',
                mem_min = '512M',
                mem_max = '2G',
                debug = False,
                force_upgrade = False,
                backup = False,
                optifine = False
            )
            Game(config).startClient()
            self.infoLabel.text = ''
        else:
            self.infoLabel.text = 'username or version is empty'


class OrzMCLApp(App):
    def build(self):
        return Content()    


def run():
    OrzMCLApp().run()

if __name__ == "__main__":
    run()
